// Copyright 1997-1999 Omni Development, Inc.  All rights reserved.
//
// This software may only be used and reproduced according to the
// terms in the file OmniSourceLicense.html, which should be
// distributed with this project and can also be found at
// http://www.omnigroup.com/DeveloperResources/OmniSourceLicense.html.

#import <OmniAppKit/OAColorPalette.h>

#import <Foundation/Foundation.h>
#import <AppKit/AppKit.h>
#import <OmniBase/OmniBase.h>
#import <OmniFoundation/OmniFoundation.h>

RCS_ID("$Header: /Network/Source/CVS/OmniGroup/OmniAppKit/OAColorPalette.m,v 1.7 1999/08/05 05:08:52 bungi Exp $")

@implementation OAColorPalette

static NSDictionary *namedColorsDictionary = nil;

#define NonBuggyCharacterSet NSMutableCharacterSet

+ (void)initialize;
{
    static BOOL initialized = NO;

    [super initialize];
    if (initialized)
	return;
    initialized = YES;

    namedColorsDictionary = [[NSDictionary alloc] initWithContentsOfFile:[[NSBundle bundleForClass:self] pathForResource:@"namedColors" ofType:@"plist"]];
}

static inline double
applyGammaAndNormalize(unsigned int sample, unsigned int maxValue, double gamma)
{
    double normalizedSample;

    normalizedSample = ((double)sample / (double)maxValue);

    if (gamma == 1.0 || gamma == 0.0)
        return normalizedSample;

    return pow(normalizedSample, gamma);
}

#define MAX_HEX_TEXT_LENGTH 40

static inline unsigned int parseHexString(NSString *hexString, unsigned long long int *parsedHexValue)
{
    unsigned int hexLength;
    unichar hexText[MAX_HEX_TEXT_LENGTH];
    unichar hexDigit;
    unsigned int textIndex;
    unsigned long long int hexValue;
    unsigned int hexDigitsFound;

    hexLength = [hexString length];
    if (hexLength > MAX_HEX_TEXT_LENGTH)
        hexLength = MAX_HEX_TEXT_LENGTH;
    [hexString getCharacters:hexText range:NSMakeRange(0, hexLength)];

    textIndex = 0;
    hexValue = 0;
    hexDigitsFound = 0;

    while (textIndex < hexLength && isspace(hexText[textIndex])) {
        // Skip leading whitespace
        textIndex++;
    }

    while (textIndex < hexLength) {
        hexDigit = hexText[textIndex++];

        if (hexDigit >= '0' && hexDigit <= '9') {
            hexDigit = hexDigit - '0';
        } else if (hexDigit >= 'A' && hexDigit <= 'F') {
            hexDigit = hexDigit - 'A' + 10;
        } else if (hexDigit >= 'a' && hexDigit <= 'f') {
            hexDigit = hexDigit - 'a' + 10;
        } else if (hexDigit == 'o' || hexDigit == 'O') {
            // Some people use 'O' rather than '0'.
            hexDigit = 0;
        } else if (isspace(hexDigit)) {
            continue;
        } else {
            hexDigitsFound = 0;
            break;
        }
        hexDigitsFound++;
        hexValue <<= 4;
        hexValue |= hexDigit;
    }

    *parsedHexValue = hexValue;
    return hexDigitsFound;
}

static inline NSColor *colorForHexString(NSString *colorString, double gamma)
{
    unsigned long long int rawColor;
    unsigned int red, green, blue;
    unsigned int maskForSingleComponent;
    unsigned int bitsPerComponent;

    switch (parseHexString(colorString, &rawColor)) {
        case 3:
            bitsPerComponent = 4;
            break;
        case 6:
            bitsPerComponent = 8;
            break;
        case 9:
            bitsPerComponent = 12;
            break;
        default:
            // Malformed hex color string
            return nil;
    }
    maskForSingleComponent = (1 << bitsPerComponent) - 1;

    blue = (unsigned int)(rawColor & maskForSingleComponent);
    rawColor >>= bitsPerComponent;
    green = (unsigned int)(rawColor & maskForSingleComponent);
    rawColor >>= bitsPerComponent;
    red = (unsigned int)(rawColor & maskForSingleComponent);

    return [NSColor colorWithCalibratedRed:applyGammaAndNormalize(red, maskForSingleComponent, gamma) green:applyGammaAndNormalize(green, maskForSingleComponent, gamma) blue:applyGammaAndNormalize(blue, maskForSingleComponent, gamma) alpha:1.0];
}

static inline NSColor *colorForNamedColorString(NSString *colorString, double gamma)
{
    NSString *namedColorString;

    namedColorString = [namedColorsDictionary objectForKey:[colorString lowercaseString]];
    if (namedColorString) {
        // Found the named color, look up its hex value and return the color
        return colorForHexString(namedColorString, gamma);
    } else {
        // Named color not found
        return nil;
    }
}

+ (NSColor *)colorForString:(NSString *)colorString gamma:(double)gamma;
{
    if ([colorString hasPrefix:@"#"]) {
        NSString *hexColorString;
        NSColor *hexColor;

        // Should be a hex color string
        hexColorString = [colorString substringFromIndex:1];
        hexColor = colorForHexString(hexColorString, gamma);
        if (hexColor) {
            return hexColor;
        } else {
            // Sometimes people set their colors to "#RED"
            return colorForNamedColorString(hexColorString, gamma);
        }
    } else {
        NSColor *namedColor;

        // Try named color string first
        namedColor = colorForNamedColorString(colorString, gamma);
        if (namedColor) {
            return namedColor;
        } else {
            // Sometimes people write hex colors without a leading "#"
            return colorForHexString(colorString, gamma);
        }
    }
}

+ (NSColor *)colorForString:(NSString *)colorString;
{
    return [self colorForString:colorString gamma:1.0];
}

+ (NSString *)stringForColor:(NSColor *)color;
{
    float red, green, blue, alpha;

    // Note: alpha is ignored
    [[color colorUsingColorSpaceName:NSCalibratedRGBColorSpace] getRed:&red green:&green blue:&blue alpha:&alpha];
    return [NSString stringWithFormat:@"#%02x%02x%02x", ((int)ceil(red * 255.0)),  ((int)ceil(green * 255.0)), ((int)ceil(blue * 255.0))];
}

@end
